#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from .service_base import DEFAULT_CONFORMS_TO, ServiceBase

__all__ = [
    "DEFAULT_CONFORMS_TO",
    "ServiceBase",
]



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from . import routes
from .app import app

"""
This module imports both, the FastAPI `app` instance and the application's 
path functions from the `routes` module. 
It also sets the server's service instance and exports the application as 
the `app` module attribute.
"""

__all__ = ["app", "routes"]



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging
import os
import shlex
from abc import ABC
from typing import Optional

import fastapi
import yaml
from starlette.routing import Route

from gavicore.models import (
    Capabilities,
    ConformanceDeclaration,
    Link,
)
from gavicore.service import Service
from gavicore.util.dynimp import import_value
from wraptile.constants import ENV_VAR_SERVICE
from wraptile.exceptions import ServiceConfigException

DEFAULT_CONFORMS_TO = [
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/core",
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/ogc-process-description",
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/json",
    # "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/html",
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/oas30",
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/job-list",
    # "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/callback",
    "http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/dismiss",
]


class ServiceBase(Service, ABC):
    @classmethod
    def load(cls) -> "ServiceBase":
        """
        Load the service instance from environment variable.

        Returns:
            A configured service instance of type `ServiceBase`.

        Raises:
            ConfigException: if a server configuration error occurs.
        """
        service_spec: str | None = os.environ.get(ENV_VAR_SERVICE)
        service_args = shlex.split(service_spec) if service_spec else []
        args = []
        kwargs = {}
        for a in service_args:
            if a.startswith("-"):
                if not a.startswith("--"):
                    kv_example = "--key[=value]"
                    raise ServiceConfigException(
                        f"Service options must have the form {kv_example!r}, "
                        f"but got {a!r}."
                    )
                kv = [p.strip() for p in a[2:].split("=", maxsplit=1)]
                k = kv[0]
                k = k.replace("-", "_")
                maybe_invert = k.startswith("no_") and len(k) > 3
                if not k.isidentifier():
                    kv_example = "--key[=value]"
                    raise ServiceConfigException(
                        f"Service options must have the form {kv_example!r}, "
                        f"but got {kv[0]!r} as key, which is not an identifier."
                    )
                v = yaml.safe_load(kv[1]) if len(kv) == 2 else True
                if maybe_invert and isinstance(v, bool):
                    kwargs[k[3:]] = not v
                else:
                    kwargs[k] = v
            else:
                args.append(a)

        if not args:
            service_help = (
                f"The service must be passed in the form {'path.to.module:service'!r} "
                "either as first command-line argument or using the environment "
                f"variable {ENV_VAR_SERVICE!r}."
            )
            raise ServiceConfigException(f"Service not specified. {service_help}")

        service_ref, args = args[0], args[1:]
        try:
            service = import_value(
                service_ref,
                type=ServiceBase,  # type: ignore[type-abstract]
                name="service",
                example="path.to.module:service",
            )
        except (ValueError, TypeError) as e:
            raise ServiceConfigException(f"{e}")
        logger = logging.getLogger("uvicorn")
        logger.info(f"Created service instance of type {type(service).__name__}")
        service.configure(*args, **kwargs)
        return service

    def __init__(
        self,
        title: str,
        description: Optional[str] = None,
        conforms_to: Optional[list[str]] = None,
    ):
        self.title = title
        self.description = description
        self.conforms_to = conforms_to or DEFAULT_CONFORMS_TO
        self.logger = logging.getLogger("uvicorn")

    # noinspection PyMethodMayBeStatic
    def configure(self, *args, **kwargs):
        """Configure this service by the given positional and keyword arguments.

        The default implementation does nothing.

        Args:
            args: positional arguments.
            kwargs: keyword arguments.

        Raises:
            ConfigException: if a server configuration error occurs.
        """

    async def get_capabilities(
        self, request: fastapi.Request, **kwargs
    ) -> Capabilities:
        app: fastapi.FastAPI = request.app
        links = [self.get_self_link(request, "get_capabilities")] + [
            Link(
                href=self.get_url(request, r.path),
                title=r.name,
                rel="service",
                type=self.get_link_type(r.name),
                hreflang="en",
            )
            for r in app.routes
            if isinstance(r, Route)
        ]
        return Capabilities(
            title=self.title,
            description=self.description,
            links=links,
        )

    async def get_conformance(self, **_kwargs) -> ConformanceDeclaration:
        return ConformanceDeclaration(conformsTo=self.conforms_to)

    @classmethod
    def get_self_link(cls, request: fastapi.Request, name: str, **path_params) -> Link:
        return Link(
            href=str(request.url_for(name, **path_params)),
            rel="self",
            title=name,
            type=cls.get_link_type(name),
            hreflang="en",
        )

    @classmethod
    def get_link_type(cls, name: str) -> str:
        html_names = ("swagger_ui_html", "swagger_ui_redirect", "redoc_html")
        return "text/html" if name in html_names else "application/json"

    @classmethod
    def get_url(cls, request: fastapi.Request, path: str):
        return str(request.base_url.replace(path=path))



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from .airflow_service import DEFAULT_AIRFLOW_BASE_URL, AirflowService

service = AirflowService(
    title="Airflow Service",
    description=(
        "A gateway API compliant with OGC API - Processes that uses an Airflow backend."
    ),
)

__all__ = [
    "DEFAULT_AIRFLOW_BASE_URL",
    "AirflowService",
    "service",
]



# generated by gen_server.py:
#   filename:  routes.py:
#   timestamp: 2025-10-01T16:12:51.474920

import fastapi

from gavicore.models import (
    ProcessRequest,
)
from gavicore.service import Service

from .app import app
from .provider import get_service


# noinspection PyPep8Naming
@app.get("/")
async def get_capabilities(
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_capabilities(request=request, response=response)


# noinspection PyPep8Naming
@app.get("/conformance")
async def get_conformance(
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_conformance(request=request, response=response)


# noinspection PyPep8Naming
@app.get("/processes")
async def get_processes(
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_processes(request=request, response=response)


# noinspection PyPep8Naming
@app.get("/processes/{processID}")
async def get_process(
    processID: str,
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_process(
        process_id=processID, request=request, response=response
    )


# noinspection PyPep8Naming
@app.post("/processes/{processID}/execution", status_code=201)
async def execute_process(
    processID: str,
    process_request: ProcessRequest,
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.execute_process(
        process_id=processID,
        process_request=process_request,
        request=request,
        response=response,
    )


# noinspection PyPep8Naming
@app.get("/jobs")
async def get_jobs(
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_jobs(request=request, response=response)


# noinspection PyPep8Naming
@app.get("/jobs/{jobId}")
async def get_job(
    jobId: str,
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_job(job_id=jobId, request=request, response=response)


# noinspection PyPep8Naming
@app.delete("/jobs/{jobId}")
async def dismiss_job(
    jobId: str,
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.dismiss_job(job_id=jobId, request=request, response=response)


# noinspection PyPep8Naming
@app.get("/jobs/{jobId}/results")
async def get_job_results(
    jobId: str,
    request: fastapi.Request,
    response: fastapi.Response,
    service: Service = fastapi.Depends(get_service),
):
    return await service.get_job_results(
        job_id=jobId, request=request, response=response
    )



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.process import ProcessPoolExecutor
from typing import Callable, Optional

import fastapi
import pydantic

from gavicore.models import (
    JobInfo,
    JobList,
    JobResults,
    JobStatus,
    ProcessDescription,
    ProcessList,
    ProcessRequest,
    ProcessSummary,
)
from procodile import Job, Process, ProcessRegistry
from wraptile.exceptions import ServiceException
from wraptile.services.base import ServiceBase


class LocalService(ServiceBase):
    def __init__(
        self,
        title: str,
        description: Optional[str] = None,
        process_registry: Optional[ProcessRegistry] = None,
    ):
        super().__init__(title=title, description=description)
        self.executor: Optional[ThreadPoolExecutor | ProcessPoolExecutor] = None
        self.process_registry = process_registry or ProcessRegistry()
        self.jobs: dict[str, Job] = {}

    def configure(
        self, processes: Optional[bool] = None, max_workers: Optional[int] = None
    ):
        """
        Configure the local service.

        Args:
            processes: Whether to use processes instead of threads. Defaults to threads.
            max_workers: The maximum number of processes or threads. Defaults to 3.
        """
        num_workers: int = max_workers or 3
        if processes:
            self.executor = ProcessPoolExecutor(max_workers=num_workers)
            self.logger.info(f"Using processes with max {num_workers} workers.")
        else:
            self.executor = ThreadPoolExecutor(max_workers=num_workers)
            self.logger.info(f"Using threads with max {num_workers} workers.")

    async def get_processes(self, request: fastapi.Request, **_kwargs) -> ProcessList:
        return ProcessList(
            processes=[
                ProcessSummary(
                    **p.description.model_dump(
                        mode="python",
                        exclude={"inputs", "outputs"},
                    )
                )
                for p in self.process_registry.values()
            ],
            links=[self.get_self_link(request, "get_processes")],
        )

    async def get_process(self, process_id: str, **kwargs) -> ProcessDescription:
        process = self._get_process(process_id)
        return process.description

    async def execute_process(
        self, process_id: str, process_request: ProcessRequest, **_kwargs
    ) -> JobInfo:
        process = self._get_process(process_id)
        job_id = f"job_{len(self.jobs)}"
        try:
            job = Job.create(process, process_request, job_id=job_id)
        except pydantic.ValidationError as e:
            raise ServiceException(
                400,
                detail=f"Invalid parameterization for process {process_id!r}: {e}",
                exception=e,
            )
        self.jobs[job_id] = job
        assert self.executor is not None, "illegal state: no executor specified"
        job.future = self.executor.submit(job.run)
        return job.job_info

    async def get_jobs(self, request: fastapi.Request, **_kwargs) -> JobList:
        return JobList(
            jobs=[job.job_info for job in self.jobs.values()],
            links=[self.get_self_link(request, "get_jobs")],
        )

    async def get_job(self, job_id: str, *args, **kwargs) -> JobInfo:
        job = self._get_job(job_id, forbidden_status_codes={})
        return job.job_info

    async def dismiss_job(self, job_id: str, *args, **_kwargs) -> JobInfo:
        job = self._get_job(job_id, forbidden_status_codes={})
        if job.job_info.status in (JobStatus.accepted, JobStatus.running):
            job.cancel()
        elif job.job_info.status in (
            JobStatus.dismissed,
            JobStatus.successful,
            JobStatus.failed,
        ):
            del self.jobs[job_id]
        return job.job_info

    async def get_job_results(self, job_id: str, *args, **_kwargs) -> JobResults:
        job = self._get_job(
            job_id,
            forbidden_status_codes={
                JobStatus.accepted: "has not started yet",
                JobStatus.running: "is still running",
                JobStatus.dismissed: "has been cancelled",
                JobStatus.failed: "has failed",
            },
        )
        assert job.job_info.status == JobStatus.successful
        assert job.future is not None
        return job.future.result()

    # noinspection PyShadowingBuiltins
    def process(
        self,
        function: Optional[Callable] = None,
        /,
        *,
        id: Optional[str] = None,
        version: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        input_fields: Optional[dict[str, pydantic.fields.FieldInfo]] = None,
        output_fields: Optional[dict[str, pydantic.fields.FieldInfo]] = None,
    ) -> Callable[[Callable], Callable]:
        """
        A decorator that can be applied to a user function in order to
        register it as a process in this registry.

        The decorator can be used with or without parameters.
        """
        return self.process_registry.process(
            function,
            id=id,
            version=version,
            title=title,
            description=description,
            input_fields=input_fields,
            output_fields=output_fields,
        )

    def _get_process(self, process_id: str) -> Process:
        process = self.process_registry.get(process_id)
        if process is None:
            raise ServiceException(404, detail=f"Process {process_id!r} does not exist")
        return process

    def _get_job(
        self, job_id: str, forbidden_status_codes: dict[JobStatus, str]
    ) -> Job:
        job = self.jobs.get(job_id)
        if job is None:
            raise ServiceException(404, detail=f"Job {job_id!r} does not exist")
        message = forbidden_status_codes.get(job.job_info.status)
        if message:
            raise ServiceException(403, detail=f"Job {job_id!r} {message}")
        return job



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging
from typing import Annotated, Optional

import typer

from wraptile.constants import (
    DEFAULT_HOST,
    DEFAULT_PORT,
    ENV_VAR_SERVER_HOST,
    ENV_VAR_SERVER_PORT,
    ENV_VAR_SERVICE,
)
from wraptile.logging import LogMessageFilter

CLI_HELP = """
Wraptile is a server made for wrapping workflow orchestration 
systems with a unified restful API that should be almost compliant
with the OGC API - Processes - Part 1: Core Standard.

For details see https://ogcapi.ogc.org/processes/.

The service instance may be followed by a `--` to pass one or more 
service-specific arguments and options.

Note that the service arguments may also be given by the 
environment variable `{service_env_var}`.
""".format(
    service_env_var=ENV_VAR_SERVICE,
)


def parse_cli_service_options(
    _ctx: typer.Context, kwargs: Optional[list[str]] = None
) -> list[str]:
    import os
    import shlex

    if not kwargs:
        return []
    service_args = os.environ.get(ENV_VAR_SERVICE)
    if kwargs == [service_args]:
        return shlex.split(service_args)
    return kwargs


cli = typer.Typer(name="wraptile", help=CLI_HELP, invoke_without_command=True)

cli_host_option = typer.Option(
    envvar=ENV_VAR_SERVER_HOST,
    help="Host address.",
)
cli_port_option = typer.Option(
    envvar=ENV_VAR_SERVER_PORT,
    help="Port number.",
)
cli_service_arg = typer.Argument(
    callback=parse_cli_service_options,
    envvar=ENV_VAR_SERVICE,
    help=(
        "Service instance optionally followed by `--` to pass "
        "service-specific arguments and options. SERVICE should "
        "have the form `path.to.module:service`."
    ),
    metavar="SERVICE [-- SERVICE-OPTIONS]",
)


@cli.callback()
def main(
    _ctx: typer.Context,
    version_: Annotated[
        bool, typer.Option("--version", help="Show version and exit.")
    ] = False,
):
    if version_:
        from importlib.metadata import version

        typer.echo(version("wraptile"))
        raise typer.Exit()


@cli.command()
def run(
    host: Annotated[str, cli_host_option] = DEFAULT_HOST,
    port: Annotated[int, cli_port_option] = DEFAULT_PORT,
    service: Annotated[Optional[list[str]], cli_service_arg] = None,
):
    """Run server in production mode."""
    run_server(
        host=host,
        port=port,
        service=service,
        reload=False,
    )


@cli.command()
def dev(
    host: Annotated[str, cli_host_option] = DEFAULT_HOST,
    port: Annotated[int, cli_port_option] = DEFAULT_PORT,
    service: Annotated[Optional[list[str]], cli_service_arg] = None,
):
    """Run server in development mode."""
    run_server(
        host=host,
        port=port,
        service=service,
        reload=True,
    )


def run_server(**kwargs):
    import os
    import shlex

    import uvicorn

    service = kwargs.pop("service", None)
    if isinstance(service, list) and service:
        os.environ[ENV_VAR_SERVICE] = shlex.join(service)

    # Apply the filter to the uvicorn.access logger
    logging.getLogger("uvicorn.access").addFilter(LogMessageFilter("/jobs"))

    uvicorn.run("wraptile.main:app", **kwargs)


if __name__ == "__main__":  # pragma: no cover
    cli()



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import traceback
from http import HTTPStatus
from typing import Optional

from fastapi import HTTPException

from gavicore.models import ApiError


class ServiceException(HTTPException):
    """Raised if a service error occurred."""

    def __init__(
        self,
        status_code: int,
        detail: str,
        exception: Optional[Exception] = None,
    ):
        super().__init__(status_code=status_code, detail=detail)
        self.content = ApiError(
            type=type(exception).__name__ if exception is not None else "ApiError",
            status=status_code,
            title=HTTPStatus(status_code).phrase,
            detail=detail,
            traceback=(
                traceback.format_exception(
                    type(exception), exception, exception.__traceback__
                )
                if exception is not None
                else None
            ),
        )


class ServiceConfigException(ServiceException):
    """Raised if a service configuration error occurred."""

    def __init__(self, message: str):
        super().__init__(status_code=500, detail=message, exception=self)



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging
import time
from typing import Awaitable, Callable

from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse

from .exceptions import ServiceException

app = FastAPI()


@app.exception_handler(ServiceException)
async def json_http_exception_handler(
    _request: Request, exc: ServiceException
) -> JSONResponse:
    return JSONResponse(
        status_code=exc.status_code,
        content=exc.content.model_dump(
            mode="json", exclude_defaults=False, exclude_none=True, exclude_unset=True
        ),
    )


@app.middleware("http")
async def log_request_duration(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    start_time = time.perf_counter()

    # Process the request
    response = await call_next(request)

    # Calculate duration
    duration = time.perf_counter() - start_time

    # Log info (method, URL path, duration in ms)
    logger = logging.getLogger("uvicorn.request_duration")
    logger.debug(
        f"{request.method} {request.url.path} completed in {duration * 1000:.2f} ms"
    )

    return response



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import datetime
import os
from functools import cached_property
from typing import Any, Optional

import fastapi
import requests
from airflow_client.client import (
    ApiClient,
    ApiException,
    Configuration,
    DAGRunPatchBody,
    DAGRunPatchStates,
    DAGRunResponse,
    DagRunState,
    TriggerDAGRunPostBody,
)
from airflow_client.client.api import DAGApi
from airflow_client.client.api import DagRunApi as DAGRunApi
from airflow_client.client.api import XComApi

from gavicore.models import (
    InputDescription,
    JobInfo,
    JobList,
    JobResults,
    JobStatus,
    JobType,
    OutputDescription,
    ProcessDescription,
    ProcessList,
    ProcessRequest,
    ProcessSummary,
)
from wraptile.exceptions import ServiceException
from wraptile.services.base import ServiceBase

DEFAULT_AIRFLOW_BASE_URL = "http://localhost:8080"


class AirflowService(ServiceBase):
    def __init__(
        self,
        title: str,
        description: Optional[str] = None,
    ):
        super().__init__(title=title, description=description)
        self._exec_count: int = 0
        self._airflow_base_url: Optional[str] = None
        self._airflow_username: Optional[str] = None
        self._airflow_password: Optional[str] = None

    def configure(
        self,
        airflow_base_url: Optional[str] = None,
        airflow_username: Optional[str] = None,
        airflow_password: Optional[str] = None,
    ):
        """
        Configure the Airflow service.

        Args:
            airflow_base_url: The base URL of the Airflow web API, defaults to
                `http://localhost:8080`.
            airflow_username: The Airflow username, defaults to `admin`.
            airflow_password: The Airflow password.
                For an Airflow installation with the simple Auth manager,
                use the one from
                `.airflow/simple_auth_manager_passwords.json.generated`.
        """
        self._airflow_base_url = airflow_base_url
        self._airflow_username = airflow_username
        self._airflow_password = airflow_password

    async def get_processes(
        self, request: fastapi.Request, *args, **kwargs
    ) -> ProcessList:
        processes: list[ProcessSummary] = []
        try:
            dag_collection = self.airflow_dag_api.get_dags(
                exclude_stale=True,
                limit=None,
                owners=None,  # TODO important, get for current user only
            )
        except ApiException as e:
            raise ServiceException(e.status, detail=e.reason, exception=e) from e
        if dag_collection and dag_collection.dags:
            for dag in dag_collection.dags:
                # https://github.com/apache/airflow-client-python/blob/main/airflow_client/client/models/dag_response.py
                processes.append(
                    ProcessSummary(
                        id=dag.dag_id,
                        version="0.0.0",  # TODO: get version
                        title=dag.dag_display_name,
                        description=dag.description,
                    )
                )
        return ProcessList(
            processes=processes,
            links=[self.get_self_link(request, "get_processes")],
        )

    async def get_process(self, process_id: str, *args, **kwargs) -> ProcessDescription:
        try:
            dag_details = self.airflow_dag_api.get_dag_details(dag_id=process_id)
        except ApiException as e:
            raise ServiceException(e.status, detail=e.reason, exception=e) from e

        inputs: dict[str, InputDescription] = {}
        if dag_details.params:
            # TODO: check that 'param_value' will be an instance of 'Param' model
            for param_key, param_value in dag_details.params.items():
                # TODO: inputs[param_key] = InputDescription(param_value. ...)
                print(f"    - {param_key} = {param_value} (type {type(param_value)}):")
        else:
            print("  No parameters (params) defined for this DAG.")

        # TODO: where to get outputs from?
        outputs: dict[str, OutputDescription] = {}

        return ProcessDescription(
            id=dag_details.dag_id,
            version="0.0.0",  # TODO: get version
            title=dag_details.dag_display_name,
            description=dag_details.doc_md or dag_details.description,
            inputs=inputs,
            outputs=outputs,
        )

    async def execute_process(
        self, process_id: str, process_request: ProcessRequest, *args, **kwargs
    ) -> JobInfo:
        logical_date = datetime.datetime.now(datetime.timezone.utc)
        dag_run_id = self.new_dag_run_id(process_id, logical_date)
        dag_run_body = TriggerDAGRunPostBody(
            dag_run_id=dag_run_id,
            conf=process_request.inputs,
            logical_date=logical_date,
        )
        try:
            dag_run = self.airflow_dag_run_api.trigger_dag_run(process_id, dag_run_body)
        except ApiException as e:
            raise ServiceException(e.status, e.reason, exception=e) from e
        return self.dag_run_to_job_info(dag_run)

    async def get_jobs(self, request: fastapi.Request, *args, **kwargs) -> JobList:
        try:
            dag_collection = self.airflow_dag_api.get_dags(
                exclude_stale=True,
                limit=None,
                owners=None,  # TODO important, get for current user only
            )
        except ApiException as e:
            raise ServiceException(e.status, detail=e.reason, exception=e) from e

        jobs: list[JobInfo] = []
        for dag in dag_collection.dags:
            try:
                dag_run_collection = self.airflow_dag_run_api.get_dag_runs(dag.dag_id)
            except ApiException as e:
                raise ServiceException(e.status, e.reason, exception=e) from e
            jobs.extend(
                self.dag_run_to_job_info(dag_run)
                for dag_run in dag_run_collection.dag_runs
            )
        return JobList(jobs=jobs, links=[self.get_self_link(request, name="get_jobs")])

    async def get_job(self, job_id: str, *args, **kwargs) -> JobInfo:
        dag_id = self.get_dag_id_from_job_id(job_id)
        try:
            dag_run = self.airflow_dag_run_api.get_dag_run(dag_id, job_id)
        except ApiException as e:
            raise ServiceException(e.status, e.reason, exception=e) from e
        return self.dag_run_to_job_info(dag_run)

    async def dismiss_job(self, job_id: str, *args, **kwargs) -> JobInfo:
        dag_id = self.get_dag_id_from_job_id(job_id)
        dag_run_patch = DAGRunPatchBody(
            note="Cancelled", state=DAGRunPatchStates.FAILED
        )
        try:
            # TODO: check if this really works
            dag_run = self.airflow_dag_run_api.patch_dag_run(
                dag_id, job_id, dag_run_patch
            )
        except ApiException as e:
            raise ServiceException(e.status, e.reason, exception=e) from e
        return self.dag_run_to_job_info(dag_run)

    async def get_job_results(self, job_id: str, *args, **kwargs) -> JobResults:
        dag_id = self.get_dag_id_from_job_id(job_id)
        return_value: Optional[Any] = None
        try:
            xcom_entry = self.airflow_xcom_api.get_xcom_entry(
                dag_id=dag_id,
                dag_run_id=job_id,
                task_id=dag_id + "_task",
                xcom_key="return_value",
            )
            if xcom_entry is not None and xcom_entry.actual_instance is not None:
                return_value = xcom_entry.actual_instance.value
        except ApiException as e:
            raise ServiceException(e.status, e.reason, exception=e) from e
        # TODO: use keys from output definitions, if provided.
        #   Otherwise, "return_value" is correct.
        return JobResults({"return_value": return_value})  # type: ignore[dict-item]

    def new_dag_run_id(self, dag_id: str, logical_time: datetime.datetime):
        self._exec_count += 1
        return f"{dag_id}__{logical_time.strftime('%Y%m%d%H%M%S')}_{self._exec_count}"

    @classmethod
    def get_dag_id_from_job_id(cls, job_id):
        return job_id.split("__", maxsplit=1)[0]

    @classmethod
    def dag_run_to_job_info(cls, dag_run: DAGRunResponse) -> JobInfo:
        return JobInfo(
            type=JobType.process,
            processID=dag_run.dag_id,
            jobID=dag_run.dag_run_id,
            status=cls.dag_run_state_to_job_status(dag_run.state),
            progress=None,  # check, it seems we have no good option here
            message=dag_run.note,  # check
            created=dag_run.queued_at,
            started=dag_run.start_date,
            updated=dag_run.last_scheduling_decision,
            finished=dag_run.end_date,
        )

    @classmethod
    def dag_run_state_to_job_status(cls, dag_run_state: DagRunState) -> JobStatus:
        mapping = {
            DagRunState.RUNNING: JobStatus.running,
            DagRunState.FAILED: JobStatus.failed,
            DagRunState.QUEUED: JobStatus.accepted,
            DagRunState.SUCCESS: JobStatus.successful,
        }
        return mapping[dag_run_state]

    @cached_property
    def airflow_dag_api(self) -> DAGApi:
        return DAGApi(self.airflow_client)

    @cached_property
    def airflow_dag_run_api(self) -> DAGRunApi:
        return DAGRunApi(self.airflow_client)

    @cached_property
    def airflow_xcom_api(self) -> XComApi:
        return XComApi(self.airflow_client)

    @cached_property
    def airflow_client(self) -> ApiClient:
        airflow_base_url: str = (
            self._airflow_base_url
            or os.getenv("AIRFLOW_API_BASE_URL")
            or DEFAULT_AIRFLOW_BASE_URL
        )
        airflow_username: str = (
            self._airflow_username or os.getenv("AIRFLOW_USERNAME") or "admin"
        )
        airflow_password = self._airflow_password or os.getenv("AIRFLOW_PASSWORD")
        if not airflow_password:
            raise RuntimeError(
                "missing Airflow password; please set env var AIRFLOW_PASSWORD"
            )
        access_token = self.fetch_access_token(
            airflow_base_url,
            username=airflow_username,
            password=airflow_password,
        )
        configuration = Configuration(host=airflow_base_url, access_token=access_token)
        return ApiClient(configuration)

    @classmethod
    def fetch_access_token(
        cls,
        base_url: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> str:
        response = requests.post(
            f"{base_url}/auth/token",
            json={"username": username, "password": password},
        )
        try:
            response.raise_for_status()
            token_data = response.json()
            return token_data.get("access_token")
        except requests.exceptions.HTTPError as e:
            raise ServiceException(
                response.status_code, detail=response.reason, exception=e
            ) from e



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging


class LogMessageFilter(logging.Filter):
    def __init__(self, *excludes: str):
        super().__init__(f"Log message filter: {excludes}")
        self.excludes = excludes

    def filter(self, record: logging.LogRecord) -> bool:
        if record.name == "uvicorn.access":
            for exclude in self.excludes:
                if exclude in record.getMessage():
                    return False
        return True



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import datetime
import time
from pathlib import Path
from typing import Optional

import pydantic
from pydantic import Field

from procodile import JobContext
from gavicore.models import Link
from wraptile.services.local import LocalService

service = LocalService(
    title="Eozilla API Server (local dummy for testing)",
    description="Local test server implementing the OGC API - Processes 1.0 Standard",
)


@service.process(
    id="sleep_a_while",
    title="Sleep Processor",
    description=(
        "Sleeps for `duration` seconds. "
        "Fails on purpose if `fail` is `True`. "
        "Returns the effective amount of sleep in seconds."
    ),
)
def sleep_a_while(
    duration: float = 10.0,
    fail: bool = False,
) -> float:
    ctx = JobContext.get()

    t0 = time.time()
    for i in range(101):
        ctx.report_progress(progress=i)
        if fail and i == 50:
            raise RuntimeError("Woke up too early")
        time.sleep(duration / 100)
    return time.time() - t0


@service.process(
    id="primes_between",
    title="Prime Processor",
    description=(
        "Returns the list of prime numbers between a `min_val` and `max_val`."
    ),
)
def primes_between(
    min_val: int = pydantic.Field(0, ge=0),
    max_val: int = pydantic.Field(100, le=100),
) -> list[int]:
    ctx = JobContext.get()

    if max_val < 2 or max_val <= min_val:
        raise ValueError("max_val must be greater 1 and greater min_val")

    ctx.report_progress(message="Step 1: Generate sieve up to sqrt(max_val)")
    limit = int(max_val**0.5) + 1
    is_prime_small = [True] * (limit + 1)
    is_prime_small[0:2] = [False, False]
    for i in range(2, int(limit**0.5) + 1):
        if is_prime_small[i]:
            for j in range(i * i, limit + 1, i):
                is_prime_small[j] = False
    small_primes = [i for i, prime in enumerate(is_prime_small) if prime]

    ctx.report_progress(
        message="Step 2: Create the sieve for the range [min_val, max_val]"
    )
    sieve_range = max_val - min_val + 1
    is_prime = [True] * sieve_range

    for p in small_primes:
        # Find the first multiple of p in the range [min_val, max_val]
        start = max(p * p, ((min_val + p - 1) // p) * p)
        for j in range(start, max_val + 1, p):
            is_prime[j - min_val] = False

    for n in range(min_val, min(min_val + 2, max_val + 1)):
        if n < 2:
            is_prime[n - min_val] = False

    ctx.report_progress(message="Done")
    return [min_val + i for i, prime in enumerate(is_prime) if prime]


@service.process(
    id="simulate_scene",
    title="Generate scene for testing",
    description=(
        "Simulate a set scene images slices for testing. "
        "Creates an xarray dataset with `periodicity` time slices "
        "and writes it as Zarr into a temporary location. "
        "Requires installed `dask`, `xarray`, and `zarr` packages."
    ),
    input_fields={
        "var_names": Field(
            title="Variable names",
            description="Comma-separated list of variable names.",
        ),
        "bbox": Field(
            title="Bounding box",
            description="Bounding box in geographical coordinates.",
            json_schema_extra=dict(format="bbox"),
        ),
        "resolution": Field(
            title="Spatial resolution",
            description="Spatial resolution in degree.",
            ge=0.01,
            le=1.0,
        ),
        "start_date": Field(
            title="Start date",
            json_schema_extra=dict(format="date"),
        ),
        "end_date": Field(
            title="End date",
            json_schema_extra=dict(format="date"),
        ),
        "periodicity": Field(
            title="Periodicity",
            description="Size of time steps in days.",
            ge=1,
            le=10,
        ),
        "output_path": Field(
            title="Output path",
            description="Local output path or URI.",
        ),
    },
)
def simulate_scene(
    var_names: str = "a, b, c",
    bbox: tuple[float, float, float, float] = (-180, -90, 180, 90),
    resolution: float = 0.5,
    start_date: str = "2025-01-01",
    end_date: str = "2025-02-01",
    periodicity: int = 1,
    output_path: Optional[str] = None,
) -> Link:
    # dependencies only required for this operation
    import dask.array as da
    import numpy as np
    import xarray as xr

    # print(
    #     dict(
    #         var_names=var_names,
    #         bbox=bbox,
    #         resolution=resolution,
    #         start_date=start_date,
    #         end_date=end_date,
    #         periodicity=periodicity,
    #     )
    # )

    var_names_: list[str] = [name.strip() for name in var_names.split(",")]
    start_date_: datetime.date = datetime.date.fromisoformat(start_date)
    end_date_: datetime.date = datetime.date.fromisoformat(end_date)

    x1, y1, x2, y2 = bbox
    x_size = round((x2 - x1) / resolution)
    y_size = round((y2 - y1) / resolution)
    time_size = round((end_date_ - start_date_).days / periodicity)
    r05 = resolution / 2

    dataset = xr.Dataset()
    dataset.coords["lon"] = xr.DataArray(
        np.linspace(x1 + r05, x2 - r05, x_size), dims="lon"
    )
    dataset.coords["lat"] = xr.DataArray(
        np.linspace(y1 + r05, y2 - r05, y_size), dims="lat"
    )
    dataset.coords["time"] = xr.DataArray(
        np.array(
            [start_date_ + datetime.timedelta(days=days) for days in range(time_size)],
            dtype=np.datetime64,
        ),
        dims="time",
    )
    for var_name in var_names_:
        dataset[var_name] = xr.DataArray(
            da.zeros(shape=(time_size, y_size, x_size)), dims=("time", "lat", "lon")
        )

    if not output_path:
        output_path = "memory://datacube.zarr"

    dataset.to_zarr(output_path, mode="w", zarr_format=3)
    if "://" in output_path:
        href = output_path
    else:
        href = Path(output_path).resolve().as_uri()
    # noinspection PyArgumentList
    return Link(href=href, hreflang=None, type="application/zarr", rel=None)


class SceneSpec(pydantic.BaseModel):
    threshold: float
    factor: float
    # TODO: uncomment and see tests fail!
    # bbox: Optional[Bbox] = None


@service.process(
    id="return_base_model",
    title="BaseModel Test",
)
def return_base_model(
    scene_spec: SceneSpec,
) -> SceneSpec:
    return scene_spec



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from typing import Final

ENV_VAR_SERVER_HOST = "EOZILLA_SERVER_HOST"
ENV_VAR_SERVER_PORT = "EOZILLA_SERVER_PORT"
ENV_VAR_SERVICE: Final = "EOZILLA_SERVICE"

DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 8008



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging

from gavicore.service import Service

from .services.base import ServiceBase


def get_service() -> Service:
    return ServiceProvider.get_instance()


class ServiceProvider:
    _service: Service | None = None

    @classmethod
    def get_instance(cls) -> Service:
        if cls._service is None:
            cls.set_instance(ServiceBase.load())
        assert cls._service is not None
        return cls._service

    @classmethod
    def set_instance(cls, service: Service):
        cls._service = service
        logger = logging.getLogger("uvicorn")
        logger.info(f"Using service instance of type {type(service).__name__}")
